#
# Copyright 2019 - binx.io B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import webbrowser
from http.server import HTTPServer

import click

from auth0_login import fatal, setting
from auth0_login.saml.callback import SAMLAccessTokenCallbackHandler
from auth0_login.util import assert_listen_port_is_available, get_listen_port_from_url


class SAMLGetAccessTokenCommand(object):
    """
    gets a SAML response token from the SAML Provider and prints
    it to standard output.

    The request is sent to `{idp_url}/samlp/{client_id}`. The callback
    is `http://localhost:{listen_port}/saml`.

    you can configure the `listen_port`, `idp_url` and
    `client_id` in the configuration file.
    """

    def __init__(self):
        self.idp_url = setting.IDP_URL
        self.client_id = setting.CLIENT_ID
        self.tokens = {}
        self.saml_response = None

    @property
    def saml_idp_url(self):
        return f'{self.idp_url}/samlp/{self.client_id}'

    @property
    def callback_url(self):
        return f'http://localhost:{setting.LISTEN_PORT}/saml'

    @property
    def listen_port(self):
        return get_listen_port_from_url(self.callback_url)

    def set_saml_response(self, saml_response):
        self.saml_response = saml_response

    def accept_saml_response(self):
        SAMLAccessTokenCallbackHandler.callback_url = self.callback_url
        SAMLAccessTokenCallbackHandler.handler = (lambda r: self.set_saml_response(r))
        httpd = HTTPServer(('0.0.0.0', self.listen_port), SAMLAccessTokenCallbackHandler)
        httpd.handle_request()
        httpd.server_close()

    def request_authorization(self):
        webbrowser.open(self.saml_idp_url)
        self.accept_saml_response()

    def run(self):
        assert_listen_port_is_available(self.listen_port)
        self.request_authorization()
        if self.saml_response:
            print(self.saml_response)
        else:
            fatal('no SAML response retrieved')


@click.command('get-token', help=SAMLGetAccessTokenCommand.__doc__)
def get_saml_token():
    cmd = SAMLGetAccessTokenCommand()
    cmd.run()



#
# Copyright 2019 - binx.io B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging

import click

from auth0_login import fatal
from auth0_login.aws import assume_role_with_saml
from auth0_login.config import setting
from auth0_login.saml import get_saml_token


@click.group(name='saml-login', help="A command line utility to obtain SAML tokens and AWS credentials.")
@click.option('--verbose', is_flag=True, default=False, help=' for tracing purposes')
@click.option('--configuration', '-c', default="DEFAULT", help='configured in .saml-login to use')
def cli(verbose, configuration):
    logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if verbose else logging.INFO))
    setting.filename = '.saml-login'
    setting.SECTION = configuration
    if not setting.exists:
        fatal('no configuration %s found in %s', configuration, setting.filename)


cli.add_command(get_saml_token)
cli.add_command(assume_role_with_saml)

if __name__ == '__main__':
    cli()



#
# Copyright 2019 - binx.io B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from http.server import BaseHTTPRequestHandler
from urllib.parse import parse_qs


class SAMLAccessTokenCallbackHandler(BaseHTTPRequestHandler):
    @staticmethod
    def handler(response): return print(response)

    def log_message(self, fmt, *args):
        logging.debug("%s - - [%s] %s\n" %
                      (self.client_address[0],
                       self.log_date_time_string(),
                       fmt % args))

    def do_POST(self):
        length = int(self.headers.get('Content-Length'))
        body = self.rfile.read(length).decode(self.headers.get('Content-Encoding', 'utf-8'))
        logging.debug('body %s', body)
        try:
            values = parse_qs(body)
        except:
            values = {}

        if 'SAMLResponse' in values:
            SAMLAccessTokenCallbackHandler.handler(''.join(values['SAMLResponse']))
            msg = f'Received SAML response'
        else:
            msg = f'failed to read a SAMLResponse from the post body'

        self.send_response(200)
        self.send_header('Connection', 'close')
        self.send_header('Content-type', 'text/plain;utf-8')
        self.end_headers()
        self.wfile.write(msg.encode('utf-8'))



#
# Copyright 2019 - binx.io B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import logging
from http.server import BaseHTTPRequestHandler
from urllib.parse import parse_qs, urlparse

import jwt
import requests


class PKCEAccessTokenCallbackHandler(BaseHTTPRequestHandler):
    client_id = None
    verifier = None
    callback_url = None
    token_url = None
    state = None

    @staticmethod
    def handler(tokens): return print(tokens.get('access_token'))

    def log_message(self, fmt, *args):
        logging.debug("%s - - [%s] %s\n" %
                      (self.client_address[0],
                       self.log_date_time_string(),
                       fmt % args))

    def write_tokens(self, tokens):
        for token in ['id_token', 'access_token']:
            try:
                if '.' in tokens[token]:
                    self.wfile.write(json.dumps(jwt.decode(tokens[token], verify=False), indent=2).encode('utf-8'))
            except jwt.DecodeError as e:
                logging.debug(f'failed to decode {token}, {e}')

    def write_reply(self, msg, loglevel=logging.DEBUG):
        self.send_response(200)
        self.send_header('Connection', 'close')
        self.send_header('Content-type', 'text/plain;utf-8')
        self.end_headers()
        self.wfile.write(msg.encode('utf-8'))
        logging.log(loglevel, msg)
        return

    def do_GET(self):
        codes = parse_qs(urlparse(self.path).query)
        state = ''.join(codes.get('state', []))
        if state != PKCEAccessTokenCallbackHandler.state:
            msg = f'Authentication failed! expected state {PKCEAccessTokenCallbackHandler.state}, received {state}.'
            self.write_reply(msg, logging.ERROR)
            return

        if 'error' in codes:
            msg = f'Authentication failed! {codes}'
            self.write_reply(msg, logging.ERROR)
            return

        body = {
            "grant_type": "authorization_code",
            "client_id": PKCEAccessTokenCallbackHandler.client_id,
            "code_verifier": PKCEAccessTokenCallbackHandler.verifier,
            "code": codes['code'][0],
            "redirect_uri": PKCEAccessTokenCallbackHandler.callback_url,
        }
        logging.debug('obtaining access token with code, %s', body)
        response = requests.post(PKCEAccessTokenCallbackHandler.token_url, json=body)
        logging.debug('status code %d, %s', response.status_code, response.text)

        if response.status_code == 200:
            self.write_reply('Authenticated! You may close this window.\n')
            tokens = response.json()
            PKCEAccessTokenCallbackHandler.handler(tokens)
            self.write_tokens(tokens)
        else:
            self.write_reply(f'Authentication failed! {response.text}', logging.ERROR)



#
# Copyright 2019 - binx.io B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import configparser
import re
from collections import namedtuple
from os import path, chmod

from auth0_login.logging import fatal

AWSAccount = namedtuple('AWSAccount', 'number alias')


class AWSAccountConfiguration(object):
    def __init__(self):
        self.config = configparser.ConfigParser()
        self.config.read([path.expanduser(path.expandvars('~/.aws-accounts'))])
        self.accounts = self.config.defaults()

    def alias_for_account(self, account) -> str:
        """
        returns the alias for `account` from ~/.aws-accounts, or `account` if none found
        """
        return next(iter(filter(lambda k: str(self.accounts[k]) == str(account), self.accounts.keys())),
                    str(account))

    def account_for_alias(self, alias) -> str:
        """
        returns the account for `alias` from ~/.aws-accounts, or none if not found
        """
        return self.accounts.get(alias, None)

    def get_account(self, account) -> AWSAccount:
        if re.match(r'^[0-9]+$', account):
            result = AWSAccount(number=account, alias=self.alias_for_account(account))
        else:
            result = AWSAccount(number=self.account_for_alias(account), alias=account)
            if not result.number:
                fatal(f'{account} is not found in ~/.aws-accounts')
        return result


aws_accounts = AWSAccountConfiguration()



#
# Copyright 2019 - binx.io B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import configparser
from collections import namedtuple
from os import chmod, path

AWSCredentials = namedtuple('AWSCredentials', 'access_key secret_key session_token expiration')


def write_aws_credentials(credentials: AWSCredentials, profile: str):
    filename = path.expanduser(path.expandvars('~/.aws/credentials'))
    config = configparser.ConfigParser()
    config.read(filename)
    if not config.has_section(profile):
        config.add_section(profile)
    config.set(profile, 'aws_access_key_id', credentials.access_key)
    config.set(profile, 'aws_secret_access_key',  credentials.secret_key)
    if credentials.session_token:
        config.set(profile, 'aws_session_token', credentials.session_token)
    else:
        config.remove_option(profile, 'aws_session_token')
    if credentials.expiration:
        config.set(profile, 'expiration', f'{credentials.expiration}')
    else:
        config.remove_option(profile, 'expiration')
    with open(filename, 'w') as f:
        config.write(f)
    chmod(filename, 0o600)
    logging.info(f'credentials saved under AWS profile {profile}.')



# Copyright 2019 - binx.io B.V.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
"""
CLI for obtaining JWT access tokens using the PKCE flow
"""

import sys
if sys.version_info < (3,6):
    sys.exit('Sorry, Python < 3.6 is not supported')

from setuptools import find_packages, setup

dependencies = ['boto3', 'pyjwt', 'click', 'requests']

setup(
    name='auth0-login',
    version="0.4.8",
    url='https://github.com/binxio/auth0-login',
    license='Apache Software License (http://www.apache.org/licenses/LICENSE-2.0)',
    author='Mark van Holsteijn',
    author_email='mark@binx.io',
    description='CLI to obtain JWT, SAML tokens, and AWS credentials using Auth0, OAuth and SAML',
    long_description=__doc__,
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=dependencies,
    tests_require=dependencies + ['pytest'],
    test_suite='tests',
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'saml-login = auth0_login.saml.__main__:cli',
            'pkce-login = auth0_login.pkce.__main__:cli',
            'aws-console = auth0_login.aws.__main__:cli',
        ],
    },
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
